// Mars Rescue Mission Challenge
// Submission by Jeremie Allard

#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <queue>
#include <stack>

//#define FIND_SHORTEST
//#define FIND_SHEAPEST

// Original map content and size
int nx0,ny0;
char **map0;

void read_map()
{
  char buf[100];
  fgets(buf,100,stdin); nx0=atoi(buf);
  fgets(buf,100,stdin); ny0=atoi(buf);
  printf("Map size: %dx%d\n",nx0,ny0);
  map0 = new char*[ny0];
  for (int y=0;y<ny0;y++)
  {
    map0[y] = new char[nx0+3]; // nx0 chars + 0x0d + 0x0a + 0x00
    fgets(map0[y],nx0+3,stdin);
  }
}

// Expanded map size
int nx,ny;

static inline int& D(int *dist, int x, int y)
{
  return dist[y*nx+x];
}

#define DIST_WALL -1
#define DIST_UNKNOWN -2

// Compute distance from all position to the specified cell
void init_dist(int*& dist, char dest)
{
  int x,y;
  nx = 10*nx0+20;
  ny = 10*ny0+20;
  dist = new int[nx*ny];
  // First init borders
  for (y=0;y<ny;y++)
    for (x=0;x<nx;x++)
    {
      if (x<10 || y<10 || x>=nx-19 || y>=ny-19)
	D(dist,x,y)=DIST_WALL;
      else
	D(dist,x,y)=DIST_UNKNOWN;
    }
  // Then read original map
  for (int y0=0;y0<ny0;y0++)
    for (int x0=0;x0<nx0;x0++)
    {
      char c=map0[y0][x0];
      int d; 
      if (c==dest)
	d = 0;
      else if (c=='#')
	d = DIST_WALL;
      else continue; // empty space
      // fill all positions hitting this cell
      for (int y=10*y0+1;y<10*y0+20;y++)
	for (int x=10*x0+1;x<10*x0+20;x++)
	{
	  if (D(dist,x,y)!=DIST_WALL) D(dist,x,y)=d;
	}
    }
}

// Store a position
struct pos
{
  short x,y;
  pos(short _x=0, short _y=0) : x(_x), y(_y) {};
  pos operator +(const pos& p) const { return pos(x+p.x,y+p.y); }
};

static inline int& D(int* dist, const pos& p)
{
  return D(dist,p.x,p.y);
}

// 8-neighbours positions
static const pos dpos[8] =
{
  pos(0,-1), pos(-1,0), pos(1,0), pos(0,1),
  pos(-1,-1), pos(1,-1), pos(-1,1), pos(1,1),
};

// Compute distance map
void calc_dist(int* dist)
{
  int x,y;
  std::queue<pos> q;
  for (y=0;y<ny;y++)
    for (x=0;x<nx;x++)
      if(D(dist,x,y)==0)
	q.push(pos(x,y));
  printf("%d initial positions\n",q.size());
  int d0 = 0;
  while (!q.empty())
  {
    pos p0 = q.front(); q.pop();
    d0 = D(dist,p0);
    for (int i=0;i<8;i++)
    {
      pos p = p0+dpos[i];
      if (D(dist,p) == DIST_UNKNOWN)
      {
	D(dist,p) = d0+1;
	q.push(p);
      }
    }
  }
  printf("Max distance: %d\n",d0);
}

// Associate an ID to all accessible positions

int* id;
int nb_id;

void calc_id(int* dist)
{
  nb_id=0;
  id = new int[nx*ny];
  for (int y=0;y<ny;y++)
    for (int x=0;x<nx;x++)
    {
      if (D(dist,x,y)>=0)
	D(id,x,y) = nb_id++;
      else
	D(id,x,y) = -1;
    }
}

// Find a position given a cell type (to find the starting position)
pos find(char c)
{
  for (int y0=0;y0<ny0;y0++)
    for (int x0=0;x0<nx0;x0++)
      if (map0[y0][x0]==c)
	return pos(10*x0+10,10*y0+10);
  return pos(0,0);
}

// Write the shortest path given a distance map
void write_path(int* dist,pos p)
{
  int i;
  while (D(dist,p)>0)
  {
    pos p2=p;
    for (i=0;i<8;i++)
    {
      p2 = p+dpos[i];
      if (D(dist,p2) == D(dist,p)-1) break;
    }
    if (i==8) return;
    p=p2;
    int x0=(p.x-10)/10;
    int y0=(p.y-10)/10;
    if (map0[y0][x0]==' ') map0[y0][x0]='.';
  }
}

void print_map()
{
  for (int y=0;y<ny0;y++)
    printf("%s",map0[y]);
}

int *target;
int *base;

int min_dist; // min distance from hitting target to hitting base
              // gives an idea of how long the return trip will be

void find_min_dist(int* d0, int* d1)
{
  min_dist=-1;
  for (int y=0;y<ny;y++)
    for (int x=0;x<nx;x++)
    {
      if (D(d0,x,y)==0)
      {
	int d = D(d1,x,y);
	if (min_dist==-1 || d>min_dist) min_dist=d;
      }
    }
  printf("min dist: %d\n",min_dist);
}

// Store a velocity
struct vel
{
  char x,y;
  vel(char _x=0, char _y=0) : x(_x), y(_y) {};
  vel operator +(const vel& p) const { return vel(x+p.x,y+p.y); }
};

pos operator +(const pos& p, const vel& v) { return pos(p.x+v.x,p.y+v.y); }

// Game constants: gravity and engines forces
const vel gravity(0,1);
const vel left(-2,0);
const vel right(2,0);
const vel bottom(0,-2);

// All possible accelerations
const vel dvel[6] = { gravity+left, gravity+bottom, gravity+right, gravity+left+bottom, gravity+right+bottom, gravity };
// Cost of each move
const int dfuel[6] = { 1, 1, 1, 2, 2, 0 };
// Displayed character for each move
const char dvelc[6] = { '1', '2', '4', '3', '6', '0' };

// To save memory the path history is stored on disc
// For each search state we store the id of the previous state (to be able to trace the path back), and the move used

FILE* fpath;
int nb_path=0;

struct __attribute__((packed)) path_record
{
  int prev;
  char d;
};

// Cache of the last writen states

#define PATH_CACHE 4096

path_record cache[PATH_CACHE];

void init_path(const char* fname)
{
  fpath = fopen(fname,"wb+");
  nb_path=0;
}

// Add a new state and get its id
int path_write(int d, int prev)
{
  if (nb_path && !(nb_path&(PATH_CACHE-1)))
  {
    fseek(fpath,0,SEEK_END);
    fwrite(&cache,sizeof(path_record),PATH_CACHE,fpath);
  }
  cache[nb_path&(PATH_CACHE-1)].d = d;
  cache[nb_path&(PATH_CACHE-1)].prev = prev;
  return nb_path++;
}

// Print a path given a state id
void path_print(int path)
{
  path_record p;
  if (path >= nb_path-PATH_CACHE)
  {
    p = cache[path&(PATH_CACHE-1)];
  }
  else
  {
    fseek(fpath,path*sizeof(path_record),SEEK_SET);
    fread(&p,sizeof(path_record),1,fpath);
  }
  if (p.prev>=0) path_print(p.prev);
  putchar(dvelc[p.d]);
}

// In-memory state information
struct state
{
  int path;         // On-disk state ID
  pos p;            // Current position
  vel v;            // Current velocity
  short step;       // Steps used
  short fuel;       // Fuel used
  bool hit_target;  // Did we already hit the target

  // Initial state
  state (const pos& _p=pos()) : path(-1), p(_p), v(0,0), step(0), fuel(0), hit_target(false) {}

  // Advance to a new state
  state (const state* s, int d)
  : path(s->path), p(s->p), v(s->v), step(s->step), fuel(s->fuel), hit_target(s->hit_target)
  {
    v = v+dvel[d];
    fuel+=dfuel[d];
    if (v.x<-10) v.x=-10; else if (v.x>10) v.x=10;
    if (v.y<-10) v.y=-10; else if (v.y>10) v.y=10;
    while (D(target,p+v)==DIST_WALL)
    {
      v.x/=2;
      v.y/=2;
    }
    step++;
    p=p+v;
    if (!hit_target && D(target,p)==0) hit_target=true;
  }

  // Store this state's path to disc
  void save(int d)
  {
    path = path_write(d,path);
  }

  bool valid() const
  {
    return D(target,p)>=0;
  }

  // Mission complete?
  bool rescued() const
  {
    return hit_target && D(base,p)==0;
  }

  // min number of steps to completion
  int min_step() const
  {
    if (hit_target)
      return step+D(base,p)/10;
    else
      return step+D(target,p)/10+min_dist/10;
  }

  int min_step10() const
  {
    if (hit_target)
      return 10*step+D(base,p);
    else
      return 10*step+D(target,p)+min_dist+100;
  }


  // maximum number of steps to completion
  int max_step() const
  {
    if (hit_target)
      return step+D(base,p)/2;
    else
      return step+D(target,p)/2+min_dist;
  }

  void print() const
  {
    if (path>=0)
      path_print(path);
    putchar('\n');
  }

  // Get position+speed ID
  int get_id() const
  {
    return 10+v.x + 21*(10+v.y + 21*(D(id,p) + nb_id*(hit_target)));
  }

};

// Which search state is better
static inline bool operator<(const state& s1, const state& s2)
{
  // favor the one with the smaller fuel
  int v1 = s1.min_step10()*1000+s1.fuel;
  int v2 = s2.min_step10()*1000+s2.fuel;
  if (v1>v2) return true;
  else if (v1<v2) return false;
  return s1.step <= s2.step;
}

// Main algorithm

void rescue(pos start)
{
  std::priority_queue<state> q; // current states sorted in priority queue
  //  std::stack<state> q; // use this for depth-first searchs
  q.push(state(start));

  // Number of distinct states
  int nb_state = nb_id*21*21*2;
  printf ("Nb states: %d\n",nb_state);

  // Best steps/fuel to arrive to each state (to prune bad branches)
  unsigned short* best_step = new unsigned short[nb_state];
  unsigned char* best_fuel = new unsigned char[nb_state];
  memset(best_step,-1,nb_state*sizeof(unsigned short));
  memset(best_fuel,-1,nb_state*sizeof(unsigned char));

  int best_dist=min_dist+10;
  int best_min = q.top().min_step();  // Best known minimal number of steps
  int best_max = 282;                 // Best known maximal number of steps
  int max_fuel = 169;                 // Best known minimal fuel cost
  int state_explored=1;               // Number of states explored
  int max_step=0;
  bool hit_target=false;              // Did we find a path to the target yet
  int* dist = target;                 // Distance to the next objective (target or base)
  int max_size = 0;

  printf("%d: BEST MIN STEP: %d\n",state_explored,best_min);
  printf("%d: BEST MAX STEP: %d\n",state_explored,best_max);

  state fsolution;

  while (!q.empty())
  {
    if (q.size() >= max_size+100000)
    { // Display the queue size
      max_size=q.size();
      printf("%d: MAX QUEUE %d\n",state_explored,max_size);
    }
    state s = q.top(); q.pop();
    for (int i=0;i<6;i++)
    {
      state s2(&s,i);
      if (!s2.valid()) continue;

      unsigned short& last_step = best_step[s2.get_id()];
      unsigned char& last_fuel = best_fuel[s2.get_id()];

      // Condition for finding shortest solution
#ifdef FIND_SHORTEST
      if (last_step < s2.step && last_fuel < s2.fuel) continue;
#endif

      // Conditions for also finding lowest fuel solution
#ifdef FIND_SHEAPEST
      if (last_fuel < s2.fuel) continue;
      if (last_step <= s2.step && last_fuel == s2.fuel) continue;
#endif

      if (last_step > s2.step) last_step = s2.step;
      if (last_fuel > s2.fuel) last_fuel = s2.fuel;

      ++state_explored;
      if (s2.step>max_step)
      {
	max_step=s2.step;
	printf("%d: STEP %d\n",state_explored, max_step); s2.print();
      }
      int min_step = s2.min_step();
      int max_step = s2.max_step();
      if (min_step < best_min)
      {
	best_min=min_step;
	printf("%d: BEST MIN STEP: %d\n",state_explored,best_min); s2.print();
      }
      if (max_step < best_max)
      {
	best_max=max_step;
	printf("%d: BEST MAX STEP: %d\n",state_explored,best_max); s2.print();
      }
      if (!hit_target && s2.hit_target)
      { // First target hit!
	printf("%d: TARGET HIT: %d steps, %d fuel\n", state_explored, s2.step,s2.fuel); s2.print();
	hit_target=true;
	best_dist=min_dist+10;
	dist = base;
      }
      if (s2.hit_target==hit_target)
      {
	int d = D(dist,s2.p);
	if (d<best_dist)
	{ // Nearest state to completion so far
	  best_dist=d;
	  printf("%d: MIN DIST: %d\n",state_explored,best_dist); s2.print();
	}
      }
      
      if (s2.rescued())
      { // Mission complete!
#ifdef FIND_SHORTEST
	if (fsolution.step == 0 || fsolution.step > s2.step || (fsolution.step==s2.step && fsolution.fuel > s2.fuel))
#else // FIND_SHEAPEST
	if (fsolution.fuel == 0 || fsolution.fuel > s2.fuel || (fsolution.fuel==s2.fuel && fsolution.step > s2.step))
#endif
	{ // Either best fuel solution
	  printf("%d: RESCUED: %d steps, %d fuel\n", state_explored, s2.step,s2.fuel); s2.print();
	  fsolution = s2;
	}
	if (s2.fuel < max_fuel)
	  max_fuel = s2.fuel;
      }
      else if (s2.fuel <= max_fuel)
      {
	s2.save(i); // Save path to disc
	q.push(s2); // add s2 in queue
      }
    }
  }
  printf("SOLUTION: %d steps, %d fuel\n",fsolution.step,fsolution.fuel); fsolution.print();
}

#ifdef WRITE_IMG

#include <IL/il.h>

void write_dist(int* dist, char* fname)
{
  unsigned char* imgdata = new unsigned char[nx*ny*3]; // RGB raw data
  int max_dist = 0;
  for (int i=0;i<nx*ny;i++)
    if (max_dist < dist[i]) max_dist=dist[i];

  for (int y=0;y<ny;y++)
  {
    unsigned char* dest = imgdata+(ny-1-y)*(nx*3); // image is upside-down
    for (int x=0;x<nx;x++)
    {
      int d = *dist;
      if (d == DIST_WALL)
      { // Walls -> RED
	dest[0] = 128;
	dest[1] = 0;
	dest[2] = 0;
      }
      else if (d < 0)
      { // Other cells -> blue
	dest[0] = 0;
	dest[1] = 0;
	dest[2] = 255;
      }
      else if ((d%50)==0)
      {
	dest[0] = 255;
	dest[1] = 255;
	dest[2] = 255;
      }
      else
      {
	unsigned char c = (unsigned char)(255-(d*255)/max_dist);
	dest[0] = 0;
	dest[1] = c;
	dest[2] = 0;
      }
      dest+=3;
      ++dist;
    }
}
  // Use DevIL to save the image

  printf("Writing image %s\n",fname);

  //ilDefaultImage();

  ilTexImage(nx, ny, 1, 3, IL_RGB, IL_UNSIGNED_BYTE, imgdata);

  ilSaveImage(fname);

  delete[] imgdata;

}
#endif

int main(int argc, char** argv)
{
  read_map();
  init_dist(target,'+');
  calc_dist(target);
  init_dist(base,'O');
  calc_dist(base);

#ifdef WRITE_IMG
  ilInit();
  write_dist(target,"target.png");
  write_dist(base,"base.png");
  return 0;
#endif
  pos pbase = find('O');
  pos ptarget = find('+');
  printf("Base: <%d,%d> distance=%d\n",pbase.x,pbase.y,D(target,pbase));
  printf("Target: <%d,%d> distance=%d\n",ptarget.x,ptarget.y,D(base,ptarget));
  write_path(target,pbase);
  print_map();
  find_min_dist(target,base);
  calc_id(target);
  init_path((argc<2)?"path.tmp":argv[1]);
  rescue(pbase);
  return 0;
}
